@@ -42,10 +42,10 @@ COMPILED_MISSING_PATTERNS = re.compile('^%s$' % '|'.join(MISSING_PATTERNS.values
 # config-y stuff in it that might be useful outside of that function alone
 LONG_MAPS = {
     'CHGS': {
-        'CHARGE': ['CHG', 'REVCHG'],
-        'RATE': ['RATE'],
-        'REVCODE': ['REVCD'],
-        'UNITS': ['UNIT'],
+        'CHARGE': ['CHG', 'REVCHG', 'CHRGS_LINE_ITEM'],
+        'RATE': ['RATE', 'UNIT_RATE'],
+        'REVCODE': ['REVCD', 'REVENUE_CODE'],
+        'UNITS': ['UNIT', 'UNITS_OF_SERVICE'],
     },
     'DX': {
         'DX': ['DX'],
@@ -75,6 +75,31 @@ LONG_MAPS = {
 }
 
 
+def lm_reverse(category, entry):
+    """
+    Finds the matching "column" for a given entry value, where
+    entry is a string that appears in a long map list of possible
+    columns.
+    """
+    lm = LONG_MAPS
+    
+    matched = False
+    for col, candidates in lm[category].iteritems():
+        if matched == False:
+        
+            if (
+                entry.lower() == col.lower() or # matches straight-up
+                entry.lower() in [x.lower() for x in candidates]
+                ):
+                matched = True
+                match = col
+    
+    if matched:
+        return match
+    else:
+        return False
+
+
 def read(target, meta_df=None, state=None, year=None, variety=None,
          skiprows=None, nrows=None, chunksize=None):
     """Read in target data file. Uses supplied meta_df or infers from state, year, variety.
@@ -135,47 +160,156 @@ def replace_df_placeholders(df, extra_dict=None):
 
 
 class UtilizationFlagger:
-    def __init__(self, uflag_name, revcodes=None, pr=None, prccs=None):
+    def __init__(self, uflag_name, revcodes=None, chg_group=None, pr=None, prccs=None):
         """Object holding a definition for a utilization flag. Can also apply the definition to a set of charges and procedures data in order to determine the appropriate value for the utilization flag. Must specify at least one of revcodes, pr, and prccs.
         
         Parameters
         revcodes: list (optional)
-            Contains a list of UB92/UB04 revenue codes associated with the utilitzation flag. Must contain codes as integers
+            Contains a list of UB92/UB04 revenue codes associated with the utilitzation flag.
+        
+        chg_group: list (optional)
+            Contains a list of charge group numbers, for state-years with charge buckets rather than per-revenue code charges. These will correspond to the n in CHGn in the source files.
+            
+            WARNING: If revcodes is set, chg_group is ignored.
         
         pr: list (optional)
-            Contains a list of ICD-9-CM procedure codes associated with the utilitzation flag. Must contain codes as integers.
+            Contains a list of ICD-9-CM procedure codes associated with the utilitzation flag.
         
         prccs: list (optional)
-            Contains a list of Clinical Classification Software (CCS) procedure codes associated with the utilization flag. Must contain codes as integers.
+            Contains a list of Clinical Classification Software (CCS) procedure codes associated with the utilization flag.
         """
         # save passed parameters
         self.name = uflag_name
         self.revcodes = revcodes
+        self.chg_group = chg_group
         self.pr = pr
         self.prccs = prccs
         
         # validation of passed parameters
         if (
+            not isinstance(prccs, list) and
+            not isinstance(chg_group, list) and
             not isinstance(revcodes, list) and
-            not isinstance(pr, list) and
-            not isinstance(prccs, list)
+            not isinstance(pr, list)
             ):
-            raise Exception("At least one of revcodes, pr, and prccs must be a list")
+            raise Exception("At least one of revcodes, chg_group, pr, and prccs must be a list")
         
         if not isinstance(uflag_name, str):
             raise Exception("uflag_name must be a string (got %s)" % type(uflag_name))
         
-        for v in ['revcodes', 'pr', 'prccs']:
+        for v in ['revcodes', 'chg_group', 'pr', 'prccs']:
             a = getattr(self, v)
             if a is not None and not isinstance(a, list):
                 raise Exception("%s must be None or a list" % v)
             elif isinstance(a, list):
-                try:
-                    setattr(self, '_' + v, [int(i) for i in a] + [str(i) for i in a])
-                except:
-                    raise Exception("If %s is a list, it can only contain integers" % v)
+                # set an internal attribute version with left-filled versions included
+                setattr(self, '_' + v, [str(i).zfill(4) for i in a if str(i).zfill(4) not in [str(j) for j in a]] + [str(i) for i in a])
             else:
-                setattr(self, '_' + v, [])
+                setattr(self, '_' + v, None)
+    
+    
+    def sql(self, tbl_out, tbl_core, tbl_pr, tbl_charges):
+        """Generates a SQL statement for generating values for this selflag within a database.
+        """
+        sel_cols = ['t_core.key']
+        
+        # revcode part
+        if self._revcodes is not None:
+            # have some revcodes to check for
+            rev_clause = """
+            CASE WHEN (
+                SUM(CASE WHEN t_revcode.revcode IN ({cs_revcodes}) AND t_revcode.CHARGE > 0 THEN 1 ELSE 0 END)
+                ) > 0
+                THEN true
+                ELSE false
+                END AS has_revcode
+            """.format(cs_revcodes=', '.join("'%s'" % x for x in self._revcodes))
+        elif self._chg_group is not None:
+            # no revcodes, but we do have charge group_number values to check for
+            rev_clause = """
+            CASE WHEN (
+                SUM(CASE WHEN t_revcode.group_number IN ({cs_chg_group}) AND t_revcode.CHARGE > 0 THEN 1 ELSE 0 END)
+                ) > 0
+                THEN true
+                ELSE false
+                END AS has_revcode
+            """.format(cs_chg_group=', '.join("'%s'" % x for x in self._chg_group))
+        else:
+            rev_clause = """
+            false AS has_revcode
+            """
+        
+        sel_cols.append(rev_clause)
+        
+        # prcode part
+        if self._pr is not None and self._prccs is not None:
+            # have both cpt pr codes and also ccs pr codes
+            pr_clause = """
+            CASE WHEN (
+                SUM(CASE WHEN t_pr.pr IN ({cs_prcodes}) OR t_pr.prccs IN ({cs_prccscodes}) THEN 1 ELSE 0 END)
+                ) > 0
+                THEN true
+                ELSE false
+                END AS has_prcode
+            """.format(cs_prcodes=', '.join("'%s'" % x for x in self._pr),
+                       cs_prccscodes=', '.join("'%s'" % x for x in self._prccs))
+        elif self._pr is not None:
+            # have only cpt pr codes
+            pr_clause = """
+            CASE WHEN (
+                SUM(CASE WHEN t_pr.pr IN ({cs_prcodes}) THEN 1 ELSE 0 END)
+                ) > 0
+                THEN true
+                ELSE false
+                END AS has_prcode
+            """.format(cs_prcodes=', '.join("'%s'" % x for x in self._pr))
+        elif self._prccs is not None:
+            # have only ccs pr codes
+            pr_clause = """
+            CASE WHEN (
+                SUM(CASE WHEN t_pr.prccs IN ({cs_prccscodes}) THEN 1 ELSE 0 END)
+                ) > 0
+                THEN true
+                ELSE false
+                END AS has_prcode
+            """.format(cs_prccscodes=', '.join("'%s'" % x for x in self._prccs))
+        else:
+            # have neither!
+            pr_clause = """
+            false AS has_prcode
+            """
+        
+        sel_cols.append(pr_clause)
+        
+        selflag_sql = """
+            SELECT {cs_sel_clauses}
+              FROM {tbl_core} AS t_core
+              LEFT JOIN {tbl_charges} AS t_revcode
+                ON t_core.key = t_revcode.key
+              LEFT JOIN {tbl_pr} AS t_pr
+                ON t_core.key = t_pr.key
+              GROUP BY t_core.key
+            """.format(cs_sel_clauses=",\t".join(sel_cols),
+                       tbl_core=tbl_core,
+                       tbl_charges=tbl_charges,
+                       tbl_pr=tbl_pr,
+                       )
+        
+        full_sql = """
+            WITH summary AS ({subq})
+            INSERT INTO {tbl_out} (key, name, value)
+            (SELECT key,
+            '{selflag_name}' AS selflag_name,
+            CASE
+                WHEN has_revcode AND has_prcode THEN 3
+                WHEN has_prcode THEN 2
+                WHEN has_revcode THEN 1
+                ELSE 0
+            END AS selflag_value
+            FROM summary);
+            """.format(subq=selflag_sql, selflag_name=self.name, tbl_out=tbl_out)
+        
+        return full_sql
     
     
     def __repr__(self):
@@ -262,17 +396,17 @@ class UtilizationFlagger:
         return flag
 
 
-def default_uflaggers():
+def default_uflaggers(state=None, year=None):
     """Creates a list of UtilizationFlagger objects using the default utilization flag definitions bundled in PyHCUP
+    
+    The state and year parameters are optional, and can be passed to request special case uflag definitions. Generally these apply only to state-years that provide aggregate charge groups or "buckets," rather than detailed charges.
+    
+    If both state and year are passed, the bundled uflag definitions will be checked for a special column that applies. Any additional state-years added to the csv will require code in this class function as well.
     """
     definitions = pd.read_csv(BUNDLED_UFLAGDEF)
     
     flaggers = []
     for k, d in definitions.iterrows():
-        if pd.isnull(d['revcodes']):
-            r = None
-        else:
-            r = d['revcodes'].split(';')
         
         if pd.isnull(d['pr']):
             pr = None
@@ -284,8 +418,58 @@ def default_uflaggers():
         else:
             prccs = d['prccs'].split(';')
         
+        if state is not None:
+            s_up = state.upper()
+        else:
+            s_up = None
         
-        f = UtilizationFlagger(d['name'], revcodes=r, pr=pr, prccs=prccs)
+        if s_up == 'AZ':
+            # Arizona special snowflakes
+            if year > 2007 and not pd.isnull(d['chgn_az_gt_2007']):
+                chg_group = d['chgn_az_gt_2007'].split(';')
+            elif year > 1994 and not pd.isnull(d['chgn_az_gt_1994']):
+                chg_group = d['chgn_az_gt_1994'].split(';')
+            else:
+                chg_group = None
+        
+        elif s_up == 'FL':
+            # Branch Floridians
+            if year > 2009 and not pd.isnull(d['chgn_fl_gt_2009']):
+                chg_group = d['chgn_fl_gt_2009'].split(';')
+            elif year > 2006 and not pd.isnull(d['chgn_fl_gt_2006']):
+                chg_group = d['chgn_fl_gt_2006'].split(';')
+            elif year > 1991 and not pd.isnull(d['chgn_fl_gt_1991']):
+                chg_group = d['chgn_fl_gt_1991'].split(';')
+            else:
+                chg_group = None
+        
+        elif s_up == 'RI':
+            # Colossal
+            chg_group = d['chgn_ri'].split(';')
+        
+        elif s_up == 'SC':
+            # Branch Floridians
+            if year > 1995 and not pd.isnull(d['chgn_sc_gt_1995']):
+                chg_group = d['chgn_sc_gt_1995'].split(';')
+            else:
+                chg_group = None
+        
+        elif s_up == 'SD' and not pd.isnull(d['chgn_sd']):
+            # The one with Mount Rushmore
+            chg_group = d['chgn_sd'].split(';')
+            
+        else:
+            chg_group = None
+        
+        if pd.isnull(d['revcodes']) or chg_group is not None:
+            # set revcodes to None if chg_group is not None
+            r = None
+        else:
+            r = d['revcodes'].split(';')
+            
+        
+        
+        f = UtilizationFlagger(d['name'], revcodes=r, chg_group=chg_group, pr=pr, prccs=prccs)
         flaggers.append(f)
     
     return flaggers
